@@ -24,7 +24,7 @@ try:
 except ImportError:
     import pycurl
 from optparse import OptionParser
-
+from sets import Set
 
 def add_repository_via_cli(line, codename, forceYes, use_ppas):   
 
@@ -53,19 +53,23 @@ def add_repository_via_cli(line, codename, forceYes, use_ppas):
             if not(forceYes):
                 print(_("Press [ENTER] to continue or ctrl-c to cancel adding it"))
                 sys.stdin.readline()
+        else:
+            if not(forceYes):
+                print(_("Unable to prompt for response.  Please run with -y"))
+                sys.exit(1)
         
-            (deb_line, file) = expand_ppa_line(line.strip(), codename)
-            deb_line = expand_http_line(deb_line, codename)
-            debsrc_line = 'deb-src' + deb_line[3:]
-            
-            # Add the key
-            short_key = ppa_info["signing_key_fingerprint"][-8:]
-            os.system("apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys %s" % short_key)
+        (deb_line, file) = expand_ppa_line(line.strip(), codename)
+        deb_line = expand_http_line(deb_line, codename)
+        debsrc_line = 'deb-src' + deb_line[3:]
 
-            # Add the PPA in sources.list.d
-            with open(file, "w") as text_file:
-                text_file.write("%s\n" % deb_line)
-                text_file.write("%s\n" % debsrc_line)
+        # Add the key
+        short_key = ppa_info["signing_key_fingerprint"][-8:]
+        os.system("apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys %s" % short_key)
+
+        # Add the PPA in sources.list.d
+        with open(file, "w") as text_file:
+            text_file.write("%s\n" % deb_line)
+            text_file.write("%s\n" % debsrc_line)
     elif line.startswith("deb "):
         with open("/etc/apt/sources.list.d/additional-repositories.list", "a") as text_file:
             text_file.write("%s\n" % expand_http_line(line, codename))
@@ -197,7 +201,7 @@ class Repository():
     def __init__(self, application, line, file, selected):
         self.application = application
         self.line = line
-        self.file = file        
+        self.file = file
         self.selected = selected
 
     def switch(self):
@@ -233,7 +237,7 @@ class Repository():
         content = readfile.read()
         readfile.close()
         content = content.replace(self.line, "")
-        with open(self.file, "w") as writefile:
+        with open(self.file, "w") as writefile:            
             writefile.write(content)
 
         # If the file no longer contains any "deb" instances, delete it as well
@@ -254,17 +258,31 @@ class Repository():
         return "<b>%s</b>\n<small><i>%s</i></small>\n<small><i>%s</i></small>" % (name, self.line, self.file)
 
     def get_repository_name(self):
-        elements = self.line.split(" ")
-        name = elements[1].replace("deb-src ", "")
-        name = name.replace("deb ", "")
-        if name.startswith("http://") or name.startswith("ftp://"):                    
-            name = name.replace("http://", "")
-            name = name.replace("ftp://", "")
-            parts = name.split("/")
-            if len(parts) > 0:
-                name = parts[0]    
-        if self.line.startswith("deb-src"):
-            name = "%s (%s)" % (name, _("Sources")) 
+        line = self.line.strip()
+        if line.startswith("deb cdrom:"):
+            name = _("CD-ROM (Installation Disc)")
+        else:
+            elements = self.line.split(" ")
+            name = elements[1].replace("deb-src ", "")
+            name = name.replace("deb ", "")
+            if name.startswith("http://") or name.startswith("ftp://"):                    
+                name = name.replace("http://", "")
+                name = name.replace("ftp://", "")
+                try:
+                    parts = name.split("/")
+                    if len(parts) > 0:
+                        name = parts[0]
+                        subparts = name.split(".")
+                        if len(subparts) > 2:
+                            if subparts[-2] != "co":
+                                name = subparts[-2].capitalize()
+                            else:
+                                name = subparts[-3].capitalize()
+                            name = name.replace("Linuxmint", "Linux Mint")
+                except:
+                    pass
+            if self.line.startswith("deb-src"):
+                name = "%s (%s)" % (name, _("Sources")) 
         return "<b>%s</b>\n<small><i>%s</i></small>\n<small><i>%s</i></small>" % (name, self.line, self.file)
 
 class ComponentToggleCheckBox(gtk.CheckButton):
@@ -357,7 +375,7 @@ class MirrorSelectionDialog(object):
     MIRROR_URL_COLUMN = 1
     MIRROR_COUNTRY_COLUMN = 2
     MIRROR_SPEED_COLUMN = 3
-    MIRROR_SPEED_BAR_COLUMN = 4
+    MIRROR_SPEED_LABEL_COLUMN = 4
     MIRROR_COUNTRY_CODE_COLUMN = 5 # invisible
     
     def __init__(self, application, ui_builder):
@@ -366,12 +384,12 @@ class MirrorSelectionDialog(object):
         
         self._dialog = ui_builder.get_object("mirror_selection_dialog")
         self._dialog.set_transient_for(application._main_window)
-
-        self._dialog.set_title(_("Select a mirror"))
         
+        self._dialog.set_title(_("Select a mirror"))
+
         self._mirrors = None
-        self._mirrors_model = gtk.ListStore(object, str, gtk.gdk.Pixbuf, float, gtk.gdk.Pixbuf, str)
-        # mirror, name, flag, speed, speed chart, country code (used to sort by flag)
+        self._mirrors_model = gtk.ListStore(object, str, gtk.gdk.Pixbuf, float, str, str)
+        # mirror, name, flag, speed, speed label, country code (used to sort by flag)
         self._treeview = ui_builder.get_object("mirrors_treeview")
         self._treeview.set_model(self._mirrors_model)
         self._treeview.set_headers_clickable(True)
@@ -388,16 +406,14 @@ class MirrorSelectionDialog(object):
         self._treeview.append_column(col)
         col.set_sort_column_id(MirrorSelectionDialog.MIRROR_URL_COLUMN)            
         
-        r = gtk.CellRendererPixbuf()
-        col = gtk.TreeViewColumn(_("Speed"), r, pixbuf = MirrorSelectionDialog.MIRROR_SPEED_BAR_COLUMN)
+        r = gtk.CellRendererText()
+        col = gtk.TreeViewColumn(_("Speed"), r, text = MirrorSelectionDialog.MIRROR_SPEED_LABEL_COLUMN)
         self._treeview.append_column(col)
         col.set_sort_column_id(MirrorSelectionDialog.MIRROR_SPEED_COLUMN)
-        col.set_min_width(int(1.1 * SPEED_PIX_WIDTH))
+        col.set_min_width(int(1.1 * SPEED_PIX_WIDTH))       
         
         self._speed_test_lock = thread.allocate_lock()
-        self._current_speed_test_index = -1
-        self._best_speed = -1
-        
+        self._meaningful_speed_threads = Set()
         self._speed_pixbufs = {}
         self.country_info = CountryInformation()
     
@@ -411,95 +427,100 @@ class MirrorSelectionDialog(object):
                 mirror,
                 mirror.url,
                 gtk.gdk.pixbuf_new_from_file(flag),
-                -1,
+                0,
                 None,
                 mirror.country_code.lower()
             ))
-        self._next_speed_test()
-    
-    def _next_speed_test(self):
-        test_mirror = None
-        for i in range(len(self._mirrors_model)):
-            url = self._mirrors_model[i][MirrorSelectionDialog.MIRROR_URL_COLUMN]
-            speed = self._mirrors_model[i][MirrorSelectionDialog.MIRROR_SPEED_COLUMN]
-            if speed == -1:
-                test_mirror = url
-                self._current_speed_test_index = i
-                break
-        if test_mirror:
-            self._speed_test_result = None
-            gobject.timeout_add(100, self._check_speed_test_done)
-            thread.start_new_thread(self._speed_test, (test_mirror,))
-    
-    def _check_speed_test_done(self):
-        self._speed_test_lock.acquire()
-        speed_test_result = self._speed_test_result
-        self._speed_test_lock.release()
-        if speed_test_result != None and len(self._mirrors_model) > 0:
-            self._mirrors_model[self._current_speed_test_index][MirrorSelectionDialog.MIRROR_SPEED_COLUMN] = speed_test_result
-            self._best_speed = max(self._best_speed, speed_test_result)
-            self._update_relative_speeds()
-            self._next_speed_test()
-            return False
-        else:
-            return True
-    
-    def _update_relative_speeds(self):
-        if self._best_speed > 0:
-            for i in range(len(self._mirrors_model)):
-                self._mirrors_model[i][MirrorSelectionDialog.MIRROR_SPEED_BAR_COLUMN] = self._get_speed_pixbuf(int(100 * self._mirrors_model[i][MirrorSelectionDialog.MIRROR_SPEED_COLUMN] / self._best_speed))
-    
-    def _get_speed_pixbuf(self, speed):
-        represented_speed = 10 * (speed / 10)
+        thread.start_new_thread(self._all_speed_tests, ())
+
+        
+    def _all_speed_tests(self):
+        self._meaningful_speed_threads.clear()
+        iter = self._mirrors_model.get_iter_first()
+        self.num_threads = 0
+        self.max_threads = 80
+        while iter is not None:
+            while self.num_threads > self.max_threads:
+                # too many threads open, let's wait... (if we open too many we can crash the app)
+                pass
+            thread_id = thread.start_new_thread(self._speed_test, (self._mirrors_model, iter))
+            self.num_threads += 1
+            self._meaningful_speed_threads.add(thread_id)
+            iter = self._mirrors_model.iter_next(iter)
+        
+    def _get_speed_label(self, speed):                
         if speed > 0:
-            if not speed in self._speed_pixbufs:
-                color_pix = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, False, 8, SPEED_PIX_WIDTH * speed / 100, SPEED_PIX_HEIGHT)
-                red = 0xff000000
-                green = 0x00ff0000
-                if represented_speed > 50:
-                    red_level = (100 - represented_speed) / 50.
-                    green_level = 1
-                else:
-                    red_level = 1
-                    green_level = (represented_speed / 50.)
-                red_level = int(255 * red_level) * 0x01000000
-                green_level = int(255 * green_level) * 0x00010000
-                color = red_level + green_level
-                color_pix.fill(color)
-                final_pix = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, False, 8, SPEED_PIX_WIDTH, SPEED_PIX_HEIGHT)
-                final_pix.fill(0xffffffff)
-                color_pix.copy_area(0, 0, SPEED_PIX_WIDTH * speed / 100, SPEED_PIX_HEIGHT, final_pix, 0, 0)
-                del color_pix
-                self._speed_pixbufs[speed] = final_pix
-            pix = self._speed_pixbufs[speed]
+            represented_speed = (speed / 1000)   # translate it to kB/S
+            unit = _("kB/s")
+            if represented_speed > 1000:
+                represented_speed = (speed / 1000)   # translate it to MB/S
+                unit = _("MB/s")
+            if represented_speed > 1000:
+                represented_speed = (speed / 1000)   # translate it to GB/S
+                unit = _("GB/s")
+            represented_speed = "%d %s" % (represented_speed, unit)            
         else:
-            pix = None
-        return pix
+            represented_speed = ("0 %s") % _("kB/s")
+        return represented_speed
     
-    def _speed_test(self, url):
+    def _speed_test(self, model, iter, depth=0):    
+        self._speed_test_lock.acquire()
+        try:
+            if iter is not None:
+                url = model.get_value(iter, MirrorSelectionDialog.MIRROR_URL_COLUMN)
+        except:
+            self._speed_test_lock.release()
+            return
+        self._speed_test_lock.release()
+
         try:
             c = pycurl.Curl()
             buff = cStringIO.StringIO()
             c.setopt(pycurl.URL, url)
-            c.setopt(pycurl.CONNECTTIMEOUT, 10)
-            c.setopt(pycurl.TIMEOUT, 10)
+            c.setopt(pycurl.CONNECTTIMEOUT, 5)
+            c.setopt(pycurl.TIMEOUT, 5)
             c.setopt(pycurl.FOLLOWLOCATION, 1)
             c.setopt(pycurl.WRITEFUNCTION, buff.write)
             c.setopt(pycurl.NOSIGNAL, 1)
             c.perform()
-            download_speed = c.getinfo(pycurl.SPEED_DOWNLOAD)
-        except:
-            download_speed = -2
-        self._speed_test_lock.acquire()
-        self._speed_test_result = download_speed
-        self._speed_test_lock.release()
+            download_speed = c.getinfo(pycurl.SPEED_DOWNLOAD) # bytes/sec
+        except pycurl.error, error:
+            errno, errstr = error            
+            if errno == 28:
+                # TIMEOUT - but since we're not launching all tests in one go, it doesn't necessarilly mean the server is timeout, maybe we're out of sockets on the client side, so let's retry
+                if (depth < 5):                    
+                    self._speed_test(model, iter, depth+1)
+                else:
+                    # Too many timeouts, removing the mirror..
+                    self._speed_test_lock.acquire()
+                    model.remove(iter)
+                    self._speed_test_lock.release()
+                return
+            else:
+                print "Error '%s' on url %s" % (errstr, url)
+                download_speed = 0
+                # Dodgy mirror, removing it...
+                try:
+                    self._speed_test_lock.acquire()
+                    if iter is not None:
+                        model.remove(iter)
+                    self._speed_test_lock.release()
+                except Exception, detail:
+                    print detail                
+
+        if thread.get_ident() in self._meaningful_speed_threads:  #otherwise, thread is "expired"
+            model.set_value(iter, MirrorSelectionDialog.MIRROR_SPEED_COLUMN, download_speed)
+            model.set_value(iter, MirrorSelectionDialog.MIRROR_SPEED_LABEL_COLUMN, self._get_speed_label(download_speed))
+
+        self.num_threads -= 1
     
-    def run(self, mirrors):
+    def run(self, mirrors):        
         self._mirrors = mirrors
-        self._best_speed = -1
         self._update_list()
         self._dialog.show_all()
-        if self._dialog.run() == gtk.RESPONSE_APPLY:
+        retval = self._dialog.run()
+        self._meaningful_speed_threads.clear()
+        if retval == gtk.RESPONSE_APPLY:
             try:
                 model, path = self._treeview.get_selection().get_selected_rows()
                 iter = model.get_iter(path[0])
@@ -565,6 +586,7 @@ class Application(object):
         self.builder.get_object("label_title_ppa").set_markup("%s" % _("PPAs"))     
         self.builder.get_object("label_title_3rd").set_markup("%s" % _("Additional repositories"))     
         self.builder.get_object("label_title_keys").set_markup("%s" % _("Authentication keys"))     
+        self.builder.get_object("label_title_maintenance").set_markup("%s" % _("Maintenance"))
 
         self.builder.get_object("label_mirrors").set_markup("<b>%s</b>" % _("Mirrors"))    
         self.builder.get_object("label_mirror_description").set_markup("%s (%s)" % (_("Main"), self.config["general"]["codename"]) )
@@ -649,9 +671,9 @@ class Application(object):
                         line = line.replace('#', '').strip()
                         selected = False
                     if line.startswith("deb"):
-                        repository = Repository(self, line.replace('#', '').strip(), source_file, selected)                    
+                        repository = Repository(self, line, source_file, selected)                    
                         if "ppa.launchpad" in line and self.config["general"]["use_ppas"] != "false":
-                            self.ppas.append(repository)                                                
+                            self.ppas.append(repository)
                         else:                        
                             self.repositories.append(repository)
             file.close() 
@@ -662,7 +684,7 @@ class Application(object):
         self._ppa_treeview.set_model(self._ppa_model)
         self._ppa_treeview.set_headers_clickable(True)
         
-        self._ppa_model.set_sort_column_id(2, gtk.SORT_DESCENDING)
+        self._ppa_model.set_sort_column_id(2, gtk.SORT_ASCENDING)
 
         r = gtk.CellRendererToggle()
         r.connect("toggled", self.ppa_toggled)        
@@ -686,7 +708,7 @@ class Application(object):
         self._repository_treeview.set_model(self._repository_model)
         self._repository_treeview.set_headers_clickable(True)
         
-        self._repository_model.set_sort_column_id(2, gtk.SORT_DESCENDING)
+        self._repository_model.set_sort_column_id(2, gtk.SORT_ASCENDING)
 
         r = gtk.CellRendererToggle()
         r.connect("toggled", self.repository_toggled)        
@@ -709,7 +731,7 @@ class Application(object):
         self._keys_treeview.set_model(self._keys_model)
         self._keys_treeview.set_headers_clickable(True)
         
-        self._keys_model.set_sort_column_id(1, gtk.SORT_DESCENDING)        
+        self._keys_model.set_sort_column_id(1, gtk.SORT_ASCENDING)        
         
         r = gtk.CellRendererText()
         col = gtk.TreeViewColumn(_("Key"), r, markup = 1)
@@ -876,7 +898,7 @@ class Application(object):
             if ppa_info["description"] is not None:
                 description = ppa_info["description"].encode("utf-8")
                 description = description.replace("<", "&lt;").replace(">", "&gt;")
-            if self.show_confirmation_dialog(self._main_window, "<b>%s</b>\n\n%s\n\n<i>%s</i>" % (line, description, str(ppa_info["web_link"])), image):                                
+            if self.show_confirm_ppa_dialog(self._main_window, "%s\n\n%s\n\n%s" % (line, description, str(ppa_info["web_link"]))):                                
                 (deb_line, file) = expand_ppa_line(line.strip(), self.config["general"]["base_codename"])
                 deb_line = expand_http_line(deb_line, self.config["general"]["base_codename"])
                 debsrc_line = 'deb-src' + deb_line[3:]
@@ -931,7 +953,7 @@ class Application(object):
         image.set_from_file("/usr/lib/linuxmint/mintSources/3rd.png")
 
         line = self.show_entry_dialog(self._main_window, _("Please enter the name of the repository you want to add:"), "deb http://packages.domain.com/ %s main" % self.config["general"]["base_codename"], image)
-        if line is not None:                                
+        if line is not None and line.strip().startswith("deb"):
             # Add the repository in sources.list.d
             with open("/etc/apt/sources.list.d/additional-repositories.list", "a") as text_file:
                 text_file.write("%s\n" % line)
@@ -998,6 +1020,34 @@ class Application(object):
             return True
         else:
             return False
+            
+    def show_confirm_ppa_dialog(self, parent, message):
+        b = gtk.TextBuffer()
+        b.set_text(message)
+        t =  gtk.TextView(b)
+        t.set_wrap_mode(gtk.WRAP_WORD)
+        s = gtk.ScrolledWindow()
+        s.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
+        s.set_shadow_type(gtk.SHADOW_OUT)
+        default_button = gtk.RESPONSE_ACCEPT
+        confirmation_button = gtk.RESPONSE_ACCEPT
+        d = gtk.Dialog(None, parent,
+                       gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
+                       (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
+                      gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
+        d.set_size_request(550, 400)
+        d.vbox.pack_start(s, True, True, 0)
+        d.set_title("")
+        s.show()
+        s.add(t)
+        t.show()
+        d.set_default_response(default_button)
+        r = d.run()        
+        d.destroy()
+        if r == confirmation_button:
+            return True
+        else:
+            return False
 
     def show_error_dialog(self, parent, message, image=None):        
         d = gtk.MessageDialog(parent,
